﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Data.SqlClient;

namespace FluentETL.Impl
{
    public class SqlDestination : SqlBase, IDestination
    {
        private SqlConnection _activeConn = null;
        private bool _identInsert;
        private bool _deleteFirst;
        private bool _truncateFirst;
        protected Action<SqlConnection> DoPreTransform { get; set; }
        protected Action<SqlConnection> DoPostTransform { get; set; }
        protected Action<SqlConnection, IRow> DoInsert { get; set; }
        protected Action<SqlConnection, IRow> DoUpdate { get; set; }
        protected Action<SqlConnection, IRow> DoDelete { get; set; }
        protected Func<IRow, RowOperation> GetRowOperation { get; set; }

        public SqlDestination()
        {
            this.GetRowOperation = x => RowOperation.Process;
        }

        public SqlDestination ToTable(string tableName)
        {
            this.CommandText = tableName;
            this.CommandType = CommandType.TableDirect;
            SqlCommand cmd = null;
            this.InsertWithCommand((conn, row) =>
            {
                if (cmd == null)
                {
                    StringBuilder sqlInsert = new StringBuilder();
                    string[] destColNames = (from c in row.Fields
                                             select c).ToArray();
                    sqlInsert.Append("INSERT INTO " + DbUtil.SqlServerEscapeTableOrColumnName(this.CommandText) + " (");
                    sqlInsert.Append(string.Join(",", destColNames));
                    sqlInsert.Append(") VALUES (");
                    string[] destColParamNames = (from c in destColNames
                                                  select "@" + c).ToArray();
                    sqlInsert.Append(string.Join(",", destColParamNames));
                    sqlInsert.Append(")");

                    cmd = conn.CreateCommand();
                    cmd.CommandText = sqlInsert.ToString();
                }
                else
                    cmd.Parameters.Clear();
                foreach (var field in row.Fields)
                {
                    object fieldValue = row[field];
                    cmd.Parameters.AddWithValue(field, fieldValue ?? DBNull.Value);
                }
                return cmd;
            });
            return this;
        }

        public SqlDestination Connection(string connString)
        {
            this.ConnectionString = connString;
            return this;
        }

        public SqlDestination Connection(SqlConnection conn)
        {
            this.ConnectionObject = conn;
            return this;
        }

        public SqlDestination PreTransform(Action<SqlConnection> action)
        {
            this.DoPreTransform = action;
            return this;
        }

        public SqlDestination PostTransform(Action<SqlConnection> action)
        {
            this.DoPostTransform = action;
            return this;
        }

        public SqlDestination InsertWithCommand(Func<SqlConnection, IRow, SqlCommand> func)
        {
            return InsertWithAction((conn, row) => { func(conn, row).ExecuteNonQuery(); });
        }

        public SqlDestination UpdateWithCommand(Func<SqlConnection, IRow, SqlCommand> func)
        {
            return UpdateWithAction((conn, row) => { func(conn, row).ExecuteNonQuery(); });
        }

        public SqlDestination DeleteWithCommand(Func<SqlConnection, IRow, SqlCommand> func)
        {
            return DeleteWithAction((conn, row) => { func(conn,row).ExecuteNonQuery(); });
        }

        public SqlDestination InsertWithAction(Action<SqlConnection, IRow> action)
        {
            this.DoInsert = action;
            return this;
        }

        public SqlDestination UpdateWithAction(Action<SqlConnection, IRow> action)
        {
            this.DoUpdate = action;
            return this;
        }

        public SqlDestination DeleteWithAction(Action<SqlConnection, IRow> action)
        {
            this.DoDelete = action;
            return this;
        }

        public SqlDestination DetermineOperation(Func<IRow, RowOperation> func)
        {
            this.GetRowOperation = func;
            return this;
        }

        public SqlDestination IdentityInsert()
        {
            _identInsert = true;
            return this;
        }

        public SqlDestination DeleteFirst()
        {
            _deleteFirst = true;
            return this;
        }

        public SqlDestination TruncteFirst()
        {
            _truncateFirst = true;
            return this;
        }

        #region IDestination Members

        void IDestination.InTransformContext(ITransform t, Action action)
        {
            base.WithConnection(conn =>
            {
                _activeConn = conn;

                if (this.DoPreTransform != null)
                    this.DoPreTransform(conn);
                if (this.CommandType == CommandType.TableDirect)
                {
                    EnsureMissingOperationsForTable(t, conn);
                }
                if (_truncateFirst && this.CommandType == System.Data.CommandType.TableDirect)
                {
                    SqlCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "TRUNCATE TABLE " + DbUtil.SqlServerEscapeTableOrColumnName(this.CommandText);
                    cmd.ExecuteNonQuery();
                }
                if (_deleteFirst && this.CommandType == System.Data.CommandType.TableDirect)
                {
                    SqlCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "DELETE FROM " + DbUtil.SqlServerEscapeTableOrColumnName(this.CommandText);
                    cmd.ExecuteNonQuery();
                }
                if (_identInsert)
                {
                    SqlCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "SET IDENTITY_INSERT " + DbUtil.SqlServerEscapeTableOrColumnName(this.CommandText) + " ON";
                    cmd.ExecuteNonQuery();
                }
                action();
                if (_identInsert)
                {
                    SqlCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "SET IDENTITY_INSERT " + DbUtil.SqlServerEscapeTableOrColumnName(this.CommandText) + " OFF";
                    cmd.ExecuteNonQuery();
                }
                if (this.DoPostTransform != null)
                    this.DoPostTransform(conn);
            });
        }

        void IDestination.Process(IRow row)
        {
            RowOperation rowOp = this.GetRowOperation(row);
            switch (rowOp)
            {
                case RowOperation.Process:
                case RowOperation.Insert:
                    this.DoInsert(_activeConn, row);
                    break;
                case RowOperation.Update:
                    this.DoUpdate(_activeConn, row);
                    break;
                case RowOperation.Delete:
                    this.DoDelete(_activeConn, row);
                    break;
            }
        }

        void IDestination.Insert(IRow row)
        {
            this.DoInsert(_activeConn, row);
        }

        void IDestination.Update(IRow row)
        {
            this.DoUpdate(_activeConn, row);
        }

        void IDestination.Delete(IRow row)
        {
            this.DoDelete(_activeConn, row);
        }

        #endregion

        private void EnsureMissingOperationsForTable(ITransform t, SqlConnection conn)
        {
            if (this.DoUpdate == null || this.DoDelete == null)
            {
                IList<string> pkFields = DbUtil.GetSqlServerPrimaryKeys(conn, this.CommandText);
                if (pkFields.Count > 0)
                {
                    if (this.DoUpdate == null)
                    {
                        IEnumerable<string> updateFields = from c in t.GetMappedDestinationFields()
                                                           where !pkFields.Contains(c)
                                                           select c;
                        StringBuilder sqlUpdate = new StringBuilder();
                        string[] setStmts = (from c in updateFields
                                             select DbUtil.SqlServerEscapeTableOrColumnName(c) +
                                                "=" +
                                                "@" + c).ToArray();
                        string[] whereConds = (from c in pkFields
                                               select DbUtil.SqlServerEscapeTableOrColumnName(c) +
                                                "=" +
                                                "@" + c).ToArray();
                        sqlUpdate.Append("UPDATE " + DbUtil.SqlServerEscapeTableOrColumnName(this.CommandText) + " SET ");
                        sqlUpdate.Append(string.Join(",", setStmts));
                        sqlUpdate.Append(" WHERE ");
                        sqlUpdate.Append(string.Join(" AND ", whereConds));

                        SqlCommand cmd = conn.CreateCommand();
                        cmd.CommandText = sqlUpdate.ToString();
                        this.UpdateWithCommand((conn2, row) =>
                        {
                            cmd.Parameters.Clear();
                            foreach (var field in updateFields)
                            {
                                object fieldValue = row[field];
                                cmd.Parameters.AddWithValue(field, fieldValue ?? DBNull.Value);
                            }
                            foreach (var field in pkFields)
                            {
                                object fieldValue = row[field];
                                cmd.Parameters.AddWithValue(field, fieldValue ?? DBNull.Value);
                            }
                            return cmd;
                        });
                    }
                    if (this.DoDelete == null)
                    {
                        StringBuilder sqlUpdate = new StringBuilder();
                        string[] whereConditions = (from c in pkFields
                                                    select DbUtil.SqlServerEscapeTableOrColumnName(c) +
                                                        "=" +
                                                        "@" + c).ToArray();
                        sqlUpdate.Append("DELETE FROM " + DbUtil.SqlServerEscapeTableOrColumnName(this.CommandText));
                        sqlUpdate.Append(" WHERE ");
                        sqlUpdate.Append(string.Join(" AND ", whereConditions));

                        SqlCommand cmd = conn.CreateCommand();
                        cmd.CommandText = sqlUpdate.ToString();
                        this.DeleteWithCommand((conn2, row) =>
                        {
                            cmd.Parameters.Clear();
                            foreach (var field in row.Fields)
                            {
                                object fieldValue = row[field];
                                cmd.Parameters.AddWithValue(field, fieldValue ?? DBNull.Value);
                            }
                            return cmd;
                        });
                    }
                }
            }
        }
    }
}